unit urlmemo;

{
TURLMemo for Delphi 2..5
Copyright (C) 1998-2K  Paul Toth <tothpaul@free.fr>
http://tothpaul.free.fr

v1.1 - 6/6/2K
 some minor bug reported by Fdric Lambelet <flambelet@bluewin.ch>
  - Brush Color for non-white Memo
  - "After Selection" invalidation
v1.2 - 7/6/2K
 an other bug reported by Marneus Calgar <ultramar@infonie.fr>
  - TURLMemo has no parent (cause TURLMemo.Change call Paint)
v1.3 - 6/12/2001
  - added web style (crHandPoint & Font Underline) by NicolasDuplat@aol.com
  - added 'www.' '.fr/', '.com/' in dectection routine
}
{
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the  GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

----------------------------------------------------------

 URLMemo highlight URL & Mails in a standard TMemo !

 Properties
 ----------
  MailTo : boolean - add "mailto:" to mails

 Events
 ------
  OnLink : OnDblClik for URL & Mails (ShellExecute by default)

-------------------------------------------------------------

 Bugs :

- if you type a space in a link, the text before the space should be black
- the text is black after a DblClick on it

}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ShellAPI;

type
  TLink = class
   Name:string;
   Rect:TRect;
   constructor Create(Aname:string; x,y,w,h:integer);
   function GotMouse(x,y:Integer):boolean;
  end;

  TLinkEvent=procedure(Sender:TObject;LinkName:string) of object;

  TURLMemo = class(TMemo)
  private
    { Dclarations prives }
    FCanvas:TCanvas;  // we need a Canvas to Redraw URLs
    FLeftPos:integer; // Horizontal ScrollBar pos
    FLinks:TList;     // keep a Links list
    FLink :integer;   // "MouseOver" link
    EOnLink:TLinkEvent;
    FMailTo:boolean;  // add "mailto:" to "name@server"
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMKeyUp(var Message: TMessage); message WM_KEYUP;
    procedure WMLbuttonUp(var Message: TMessage); message WM_LBUTTONUP;
   protected
    { Dclarations protges }
    procedure PaintWindow(DC: HDC); override;
    procedure Paint;
    procedure ClearLinks;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure DblClick; override;
    procedure Change; override;
    property Canvas: TCanvas read FCanvas;
  public
    { Dclarations publiques }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    Function FirstVisibleLine:integer;
  published
    { Dclarations publies }
    Property OnLink:TLinkEvent read EOnLink write EOnLink;
    Property MailTo:boolean read FMailTo write FMailTo default true;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MySoft', [TURLMemo]);
end;

{ TLink }
Constructor TLink.Create(AName:string; x,y,w,h:integer);
 begin
  Name:=AName;
  Rect.Left:=x;
  Rect.Top:=y;
  Rect.Right:=x+w;
  Rect.Bottom:=y+h;
 end;

Function TLink.GotMouse(x,y:integer):boolean;
 begin
  Result:=(x>=Rect.Left)and(x<=Rect.Right)
       and(y>=Rect.Top)and(y<=Rect.Bottom);
 end;

// add a Canvas (Like TCustomControl)
constructor TURLMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TControlCanvas.Create;
  TControlCanvas(FCanvas).Control := Self;
  FLeftPos:=0;
  FLinks:=TList.Create;
  FLink :=-1;
  FMailTo:=True;
end;

destructor TURLMemo.Destroy;
begin
  ClearLinks;
  FLinks.Free;
  FCanvas.Free;
  inherited Destroy;
end;

// Memo Extensions
Function TURLMemo.FirstVisibleLine:integer;
 begin
  Result:=SendMessage(Handle,em_getfirstvisibleline,0,0);
 end;

// URL Extension
procedure TURLMemo.WMPaint(var Message: TWMPaint);
 begin
  PaintHandler(Message); // will Call PaintWindow
 end;

procedure TURLMemo.WMKeyUp(var Message: TMessage);
 begin
  inherited;
  If SelText = '' Then Invalidate;
 end;

procedure TURLMemo.WMLbuttonUp(var Message: TMessage);
 begin
  inherited;
  If SelText = '' Then Invalidate;
 end;

procedure TURLMemo.PaintWindow(DC: HDC);
begin
  inherited;
  FCanvas.Handle := DC;
  try
    Paint;
  finally
    FCanvas.Handle := 0;
  end;
end;

Procedure TURLMemo.Paint;
 Const
  StopChar=[' ','[',']','(',')','<','>'];
 var
  LineSize:word;
//  s:string;
  i,x,y:integer;
  p,b,e:integer;
  s1,s2:string;

  Procedure HighLight(s:string; what:array of string);
   var
    w:word;
   begin
    for w:=0 to High(what) do begin
     x:=FLeftPos;
     s1:=s;
     p:=pos(what[w],s1);
     while p>0 do begin
      b:=p; while (b>1) and (not (s1[b-1] in Stopchar)) do dec(b);
      e:=p; while (e<length(s1)) and (not (s1[e+1] in StopChar)) do inc(e);
      s2:=copy(s1,b,e-b+1);
      inc(x,FCanvas.TextWidth(Copy(s1,1,b-1)));
      Delete(s1,1,e);
      FCanvas.TextOut(x,y,s2);
      FLinks.Add(TLink.Create(s2,x,y,FCanvas.TextWidth(s2),LineSize));
      inc(x,FCanvas.Textwidth(s2));
      p:=pos(what[w],s1);
     end;
    end;
   end;

 begin
 // setup Font
  FCanvas.Font.Assign(Font);
  FCanvas.Font.Color:=clBlue;
  FCanvas.Brush.Color:=Color; // fix 2K :)
  LineSize:=FCanvas.TextHeight('A');
 // Get Left Pos
  FLeftPos:=1-GetScrollPos(Handle,sb_horz);
 // remove previous links
  ClearLinks;
 // highligth text
  y:=1;
  HideCaret(Handle);
  for i:=FirstVisibleLine to Lines.Count-1 do begin
   HighLight(Lines[i],['@','://','www.','.fr/','.com/']);
   inc(y,LineSize);
   if y>Height then break;
  end;
  showCaret(Handle);
 end;

Procedure TURLMemo.ClearLinks;
 var
  i:integer;
 begin
  For i:=0 to FLinks.Count-1 do TLink(FLinks[i]).Free;
  FLinks.Clear;
 end;

procedure TURLMemo.MouseMove(Shift: TShiftState; X, Y: Integer);
 var
  i:integer;

  procedure UnderLine(Link:integer; Underlined:boolean);
   begin
    Canvas.Font:=Self.Font;
    Canvas.Font.Color:=clBlue;
    if Underlined then Canvas.Font.Style:=[fsUnderline] else Canvas.Font.Style:=[];
    with TLink(fLinks[Link]) do Canvas.TextRect(Rect,Rect.Left,Rect.Top,Name);
   end;

 begin
  For i:=0 to FLinks.Count-1 do
   if TLink(FLinks[i]).GotMouse(x,y) then begin
    if fLink<>i then begin
     if fLink>=0 then Underline(fLink,False);
     FLink:=i;
     Underline(fLink,True);
     Cursor:={$ifndef VER90}crHandPoint{$else}crArrow{$endif}; // no HandPoint under delphi 2
    end;
    exit;
   end;
  if FLink>=0 then begin
   Cursor:=crDefault;
   UnderLine(fLink,False);
   FLink:=-1;
  end;
 end;

procedure TURLMemo.DblClick;
 var
  s:string;
 begin
  if FLink>=0 then begin
   s:=TLink(FLinks[FLink]).Name;
   if Assigned(EOnLink) then begin
    try  EOnLink(Self,TLink(FLinks[FLink]).Name) except end;
   end else begin
    if FMailTo then begin
     if (pos('@',s)>0) then begin
      if (pos(':',s)=0) then s:='mailto:'+s;
     end else begin
      if (pos('://',s)=0) then s:='http://'+s;
     end;
    end;
    ShellExecute(Handle,nil,PChar(s),nil,nil,sw_shownormal);
   end;
  end else
   inherited DblClick;
 end;

Procedure TURLMemo.Change;
 begin
  Inherited Change;
  if (Parent<>nil) then Paint;
 end;
end.
